import guidance
from guidance import gen, models, optional, select


def test_select_reset_pos():
    model = models.Mock()
    model += "This is" + select(options=["bad", "quite bad"])
    assert str(model) in ["This isbad", "This isquite bad"]


def test_select_simple(selected_model):
    lm = selected_model
    options = ["baad I think", "bad I think", "bad"]
    lm = lm + "Scott is quite " + select(name="bad", options=options)
    assert lm["bad"] in options


def test_select_longer():
    """This tests to ensure that the grammar is extended greedily."""
    lm = models.Mock(b"<s>Scott is a very nice man.")
    lm += "Scott is a very " + select(name="text", options=["nice", "nice man."])
    assert lm["text"] == "nice man."


def test_grammar_plus_fstring():
    @guidance(stateless=True, dedent=False)
    def test(lm):
        val = 4
        lm += f"the value of {val} is best! {gen(max_tokens=1)}"
        return lm

    lm = models.Mock()
    lm += test()
    assert "{{G|" not in str(lm)


class TestRecursion:

    def test_simple_recursion(self):

        @guidance(stateless=True, dedent=False)
        def grammar(lm):
            return lm + "x" + optional(grammar())

        grammar()

    def test_mutual_recursion(self):

        @guidance(stateless=True, dedent=False)
        def grammar1(lm):
            return lm + "x" + grammar2()

        @guidance(stateless=True, dedent=False)
        def grammar2(lm):
            return lm + "y" + optional(grammar1())

        grammar1()
        grammar2()

    def test_multiple_mutual_recursion(self):
        @guidance(stateless=True, dedent=False)
        def grammar1(lm):
            return lm + "x" + grammar2()

        @guidance(stateless=True, dedent=False)
        def grammar2(lm):
            return lm + "y" + grammar3()

        @guidance(stateless=True, dedent=False)
        def grammar3(lm):
            return lm + "z" + optional(grammar1())

        grammar1()
        grammar2()
        grammar3()

    def test_branching_mutual_recursion(self):
        @guidance(stateless=True, dedent=False)
        def grammar1(lm):
            return lm + "x" + grammar2()

        @guidance(stateless=True, dedent=False)
        def grammar2(lm):
            return lm + "y" + select([grammar1(), grammar3()])

        @guidance(stateless=True, dedent=False)
        def grammar3(lm):
            return lm + "z" + optional(grammar1())

        grammar1()
        grammar2()
        grammar3()



from ._block import block


def silent():
    open_text = f"""<||_html:<div style='display: inline-block; cursor: pointer; opacity: 0.5; width: 2px; margin-left: -2px;' onClick='this.nextSibling.style.display = "inline"; this.style.display = "none"'>&caron;</div><div style='display: none;'>_||>"""
    close_text = "<||_html:</div>_||>"
    return block(opener=open_text, closer=close_text)



import os
from pathlib import Path
import multiprocessing
from itertools import takewhile
import operator
import threading
import numpy as np
import queue
import time
import tiktoken
import re

from ._vertexai import (
    VertexAICompletion,
    VertexAIInstruct,
    VertexAIChat,
    VertexAIChatEngine,
)

_image_token_pattern = re.compile(r"<\|_image:(.*)\|>")

try:
    from vertexai.language_models import (
        TextGenerationModel,
        ChatModel,
        InputOutputTextPair,
    )
    from vertexai.preview.generative_models import GenerativeModel, Content, Part, Image
    import vertexai

    # def get_chat_response(message):
    #     vertexai.init(project="PROJECT_ID", location="us-central1")
    #     model = GenerativeModel("gemini-pro")
    #     chat = model.start_chat()
    #     response = chat.send_message(message)
    #     return response.text

    # print(get_chat_response("Hello"))
    # print(get_chat_response("What are all the colors in a rainbow?"))
    # print(get_chat_response("Why does it appear when it rains?"))
    is_vertexai = True
except ModuleNotFoundError:
    is_vertexai = False

# class GeminiCompletion(VertexAICompletion):
#     def __init__(self, model, tokenizer=None, echo=True, caching=True, temperature=0.0, max_streaming_tokens=None, **kwargs):

#         if isinstance(model, str):
#             self.model_name = model
#             self.model_obj = TextGenerationModel.from_pretrained(self.model_name)

#         # Gemini does not have a public tokenizer, so we pretend it tokenizes like gpt2...
#         if tokenizer is None:
#             tokenizer = tiktoken.get_encoding("gpt2")

#         # the superclass does all the work
#         super().__init__(
#             model,
#             tokenizer=tokenizer,
#             echo=echo,
#             caching=caching,
#             temperature=temperature,
#             max_streaming_tokens=max_streaming_tokens,
#             **kwargs
#         )

# class GeminiInstruct(VertexAIInstruct):
#     def __init__(self, model, tokenizer=None, echo=True, caching=True, temperature=0.0, max_streaming_tokens=None, **kwargs):

#         if isinstance(model, str):
#             self.model_name = model
#             self.model_obj = TextGenerationModel.from_pretrained(self.model_name)

#         # Gemini does not have a public tokenizer, so we pretend it tokenizes like gpt2...
#         if tokenizer is None:
#             tokenizer = tiktoken.get_encoding("gpt2")

#         # the superclass does all the work
#         super().__init__(
#             model,
#             tokenizer=tokenizer,
#             echo=echo,
#             caching=caching,
#             temperature=temperature,
#             max_streaming_tokens=max_streaming_tokens,
#             **kwargs
#         )


class GeminiChat(VertexAIChat):
    def __init__(
        self, model, tokenizer=None, echo=True, max_streaming_tokens=None, **kwargs
    ):
        if isinstance(model, str):
            model = GenerativeModel(model)

        # Gemini does not have a public tokenizer, so we pretend it tokenizes like gpt2...
        if tokenizer is None:
            tokenizer = tiktoken.get_encoding("gpt2")

        # the superclass does all the work
        super().__init__(
            model,
            tokenizer=tokenizer,
            echo=echo,
            max_streaming_tokens=max_streaming_tokens,
            engine_class=GeminiChatEngine,
            **kwargs,
        )


class GeminiChatEngine(VertexAIChatEngine):
    def _start_chat(self, system_text, messages):
        assert (
            system_text == ""
        ), "We don't support passing system text to Gemini models (yet?)!"
        out = self.model_obj.start_chat(history=messages)
        return out

    def _start_generator(self, system_text, messages, temperature):
        # last_user_text = messages[-1]["content"]
        formated_messages = []
        for m in messages:
            raw_parts = _image_token_pattern.split(m["content"])
            parts = []
            for i in range(0, len(raw_parts), 2):

                # append the text portion
                if len(raw_parts[i]) > 0:
                    parts.append(Part.from_text(raw_parts[i]))

                # append any image
                if i + 1 < len(raw_parts):
                    parts.append(
                        Part.from_image(Image.from_bytes(self[raw_parts[i + 1]]))
                    )
            formated_messages.append(Content(role=m["role"], parts=parts))
        last_user_parts = (
            formated_messages.pop()
        )  # remove the last user stuff that goes in send_message (and not history)

        chat_session = self.model_obj.start_chat(
            history=formated_messages,
        )

        generation_config = {"temperature": temperature}
        if self.max_streaming_tokens is not None:
            generation_config["max_output_tokens"] = self.max_streaming_tokens
        generator = chat_session.send_message(
            last_user_parts, generation_config=generation_config, stream=True
        )

        for chunk in generator:
            yield chunk.candidates[0].content.parts[0].text.encode("utf8")



from guidance import char_set, one_or_more, select, string, zero_or_more
from guidance._grammar import Byte, ByteRange
from guidance._parser import EarleyCommitParser


def test_one_or_more():
    g = one_or_more("a")
    parser = EarleyCommitParser(g)
    assert parser.valid_next_bytes() == set([Byte(b"a")])
    parser.consume_byte(b"a")
    assert parser.valid_next_bytes() == set([Byte(b"a")])


def test_zero_or_more_and_one_or_more():
    g = zero_or_more("a") + one_or_more("b")
    parser = EarleyCommitParser(g)
    assert parser.valid_next_bytes() == set([Byte(b"a"), Byte(b"b")])
    parser.consume_byte(b"a")
    assert parser.valid_next_bytes() == set([Byte(b"a"), Byte(b"b")])
    parser.consume_byte(b"b")
    assert parser.valid_next_bytes() == set([Byte(b"b")])

    parser = EarleyCommitParser(g)
    assert parser.valid_next_bytes() == set([Byte(b"a"), Byte(b"b")])
    parser.consume_byte(b"b")
    assert parser.valid_next_bytes() == set([Byte(b"b")])
    parser.consume_byte(b"b")
    assert parser.valid_next_bytes() == set([Byte(b"b")])


def test_zero_or_more_and_one_or_more_mixed():
    g = zero_or_more("a") + "test" + one_or_more("b")
    parser = EarleyCommitParser(g)
    assert parser.valid_next_bytes() == set([Byte(b"a"), Byte(b"t")])
    parser.consume_byte(b"t")
    parser.consume_byte(b"e")
    parser.consume_byte(b"s")
    assert parser.valid_next_bytes() == set([Byte(b"t")])
    parser.consume_byte(b"t")
    assert parser.valid_next_bytes() == set([Byte(b"b")])


def test_select():
    g = select(["bob", "bill", "sue"])
    parser = EarleyCommitParser(g)
    assert parser.valid_next_bytes() == set([Byte(b"b"), Byte(b"s")])
    parser.consume_byte(b"s")
    assert parser.valid_next_bytes() == set([Byte(b"u")])
    parser.consume_byte(b"u")
    assert parser.valid_next_bytes() == set([Byte(b"e")])


def test_select_nested():
    g = select(["bob", "bill", select(["mark", "mary"])])
    parser = EarleyCommitParser(g)
    assert parser.valid_next_bytes() == set([Byte(b"b"), Byte(b"m")])
    parser.consume_byte(b"m")
    assert parser.valid_next_bytes() == set([Byte(b"a")])
    parser.consume_byte(b"a")
    assert parser.valid_next_bytes() == set([Byte(b"r")])
    parser.consume_byte(b"r")
    assert parser.valid_next_bytes() == set([Byte(b"k"), Byte(b"y")])


def test_select_joined():
    g = select(["bob", "bill"]) + select(["mark", "mary"])
    parser = EarleyCommitParser(g)
    assert parser.valid_next_bytes() == set([Byte(b"b")])
    parser.consume_byte(b"b")
    assert parser.valid_next_bytes() == set([Byte(b"o"), Byte(b"i")])
    parser.consume_byte(b"i")
    assert parser.valid_next_bytes() == set([Byte(b"l")])
    parser.consume_byte(b"l")
    assert parser.valid_next_bytes() == set([Byte(b"l")])
    parser.consume_byte(b"l")
    assert parser.valid_next_bytes() == set([Byte(b"m")])
    parser.consume_byte(b"m")
    assert parser.valid_next_bytes() == set([Byte(b"a")])
    parser.consume_byte(b"a")
    assert parser.valid_next_bytes() == set([Byte(b"r")])
    parser.consume_byte(b"r")
    assert parser.valid_next_bytes() == set([Byte(b"k"), Byte(b"y")])


def test_char_set():
    g = char_set("b-f")
    parser = EarleyCommitParser(g)
    assert parser.valid_next_bytes() == set([ByteRange(b"bf")])
    parser.consume_byte(b"b")


def test_byte_mask_char_set():
    g = char_set("b-f")
    parser = EarleyCommitParser(g)
    m = parser.next_byte_mask()
    for i in range(256):
        if ord(b"b") <= i <= ord(b"f"):
            assert m[i]
        else:
            assert not m[i]


def test_byte_mask_char_set2():
    g = char_set("bf")
    parser = EarleyCommitParser(g)
    m = parser.next_byte_mask()
    for i in range(256):
        if i == ord(b"b") or i == ord(b"f"):
            assert m[i]
        else:
            assert not m[i]


def test_char_set_one_or_more():
    g = one_or_more(char_set("b-f"))
    parser = EarleyCommitParser(g)
    assert parser.valid_next_bytes() == set([ByteRange(b"bf")])
    parser.consume_byte(b"b")
    assert parser.valid_next_bytes() == set([ByteRange(b"bf")])
    parser.consume_byte(b"b")
    assert parser.valid_next_bytes() == set([ByteRange(b"bf")])
    parser.consume_byte(b"f")
    assert parser.valid_next_bytes() == set([ByteRange(b"bf")])


def test_string_utf8():
    b = bytes("¶", encoding="utf8")
    g = string("¶")
    parser = EarleyCommitParser(g)
    assert parser.valid_next_bytes() == set([Byte(b[:1])])
    parser.consume_byte(b[:1])
    assert parser.valid_next_bytes() == set([Byte(b[1:])])
    parser.consume_byte(b[1:])



from .._guidance import guidance
from .._grammar import byte_range, select


@guidance(stateless=True)
def any_char_but(lm, forbidden):
    """Allows any char except those in forbidden"""
    # TODO: extend this to support utf-8 encoded multibyte unicode characters
    forb = sorted(set([ord(x) for x in forbidden]))
    start = 0
    ranges = []
    for i in forb:
        if i == 0:
            continue
        newrange = (start, i - 1)
        if newrange[0] < newrange[1]:
            ranges.append(newrange)
        start = i + 1
    if start < 127:
        ranges.append((start, 127))
    ranges = [(i.to_bytes(1, "big"), j.to_bytes(1, "big")) for i, j in ranges]
    return select([byte_range(x[0], x[1]) for x in ranges])



from ._model import Model, Instruct, Chat

# local models
from .transformers._transformers import Transformers, TransformersChat
from .llama_cpp import LlamaCpp, LlamaCppChat, MistralInstruct, MistralChat
from ._mock import Mock, MockChat

# grammarless models (we can't do constrained decoding for them)
from ._grammarless import Grammarless
from .vertexai._vertexai import (
    VertexAI,
    VertexAIChat,
    VertexAICompletion,
    VertexAIInstruct,
)
from ._azure_openai import (
    AzureOpenAI,
    AzureOpenAIChat,
    AzureOpenAICompletion,
    AzureOpenAIInstruct,
)
from ._azureai_studio import AzureAIStudioChat
from ._openai import OpenAI, OpenAIChat, OpenAIInstruct, OpenAICompletion
from ._lite_llm import LiteLLM, LiteLLMChat, LiteLLMInstruct, LiteLLMCompletion
from ._cohere import Cohere, CohereCompletion, CohereInstruct
from ._anthropic import Anthropic, AnthropicChat
from ._googleai import GoogleAI, GoogleAIChat
from ._togetherai import (
    TogetherAI,
    TogetherAIChat,
    TogetherAIInstruct,
    TogetherAICompletion,
)



import os
from ._model import Chat, Instruct
from ._openai import (
    OpenAIChatEngine,
    OpenAI,
    OpenAIInstructEngine,
    OpenAICompletionEngine,
    OpenAIEngine,
)
from .transformers._transformers import TransformersTokenizer


class TogetherAI(OpenAI):
    def __init__(
        self,
        model,
        tokenizer=None,
        echo=True,
        api_key=None,
        max_streaming_tokens=1000,
        timeout=0.5,
        compute_log_probs=False,
        engine_class=None,
        **kwargs,
    ):
        """
        Build a new TogetherAI model object that represents a model in a given state.
        """

        tokenizer = TransformersTokenizer(
            model=model, tokenizer=tokenizer, ignore_bos_token=True
        )

        # Default base_url is the together.ai endpoint
        if not "base_url" in kwargs:
            kwargs["base_url"] = "https://api.together.xyz"
        # TogetherAI uses TOGETHERAI_API_KEY env value instead of OPENAI_API_KEY
        # We pass explicitly to avoid OpenAI class complaining about a missing key
        if api_key is None:
            api_key = os.environ.get("TOGETHERAI_API_KEY", None)
        if api_key is None:
            raise Exception(
                "The api_key client option must be set either by passing api_key to the client or by setting the TOGETHERAI_API_KEY environment variable"
            )

        if engine_class is None:
            engine_map = {
                TogetherAICompletion: OpenAICompletionEngine,
                TogetherAIInstruct: OpenAIChatEngine,
                TogetherAIChat: OpenAIChatEngine,
                TogetherAI: OpenAICompletionEngine,
            }
            for k in engine_map:
                if issubclass(self.__class__, k):
                    engine_class = engine_map[k]
                    break

        super().__init__(
            model,
            tokenizer,
            echo,
            api_key,
            max_streaming_tokens,
            timeout,
            compute_log_probs,
            engine_class,
            **kwargs,
        )


class TogetherAICompletion(TogetherAI):
    pass


class TogetherAIInstruct(TogetherAI, Instruct):
    """
    Utilizes chat endpoints to simulate a single instruction query
    together.ai will format in correct prompt template for model on their end
    """

    def get_role_start(self, name):
        if name == "instruction":
            return "<|im_start|>user\n"
        else:
            raise Exception(
                f"The TogetherAIInstruct model does not know about the {name} role type!"
            )

    def get_role_end(self, name):
        if name == "instruction":
            return "<|im_end|>"
        else:
            raise Exception(
                f"The TogetherAIInstruct model does not know about the {name} role type!"
            )


class TogetherAIChat(TogetherAI, Chat):
    pass



import ast
import asyncio
import json
import inspect
import queue
import sys
import textwrap
import types

import numpy as np


class _Rewrite(ast.NodeTransformer):
    def visit_Constant(self, node):
        if isinstance(node.value, str) and node.lineno < node.end_lineno:
            self.start_counts[node.lineno - 1] += 1
            start_line = self.source_lines[node.lineno - 1]
            start_string = start_line[node.col_offset :]

            # check for literal multiline strings
            if (
                start_string.startswith("f'''")
                or start_string.startswith("'''")
                or start_string.startswith('f"""')
                or start_string.startswith('"""')
            ):

                # track our indentation level
                if self.indentation[node.lineno - 1] is None:
                    indent = start_line[: len(start_line) - len(start_line.lstrip())]
                    for i in range(node.lineno - 1, node.end_lineno):
                        self.indentation[i] = indent
                indent = self.indentation[node.lineno - 1]

                # strip indentation when it is consistent
                lines = node.value.split("\n")
                fail = False
                new_lines = []
                for i, line in enumerate(lines):
                    if (
                        i == 0
                        and (
                            self.start_counts[node.lineno - 1] > 1
                            or not start_line.endswith("\\")
                        )
                    ) or line == "":
                        new_lines.append(line)
                    elif line.startswith(indent):
                        new_lines.append(line[len(indent) :])
                    # elif (i == 0 and line.endswith("\\")) or line == "":
                    #     new_lines.append(line)
                    else:
                        fail = True
                        break
                if not fail:
                    node.value = "\n".join(new_lines)

        return node


class normalize_notebook_stdout_stderr:
    """Remaps stdout and stderr back to their normal selves from what ipykernel did to them.

    Based on: https://github.com/ipython/ipykernel/issues/795
    """

    def __enter__(self):
        normal_stdout = sys.__stdout__.fileno()
        self.restore_stdout = None
        if (
            getattr(sys.stdout, "_original_stdstream_copy", normal_stdout)
            != normal_stdout
        ):
            self.restore_stdout = sys.stdout._original_stdstream_copy
            sys.stdout._original_stdstream_copy = normal_stdout

        normal_stderr = sys.__stderr__.fileno()
        self.restore_stderr = None
        if (
            getattr(sys.stderr, "_original_stdstream_copy", normal_stderr)
            != normal_stderr
        ):
            self.restore_stderr = sys.stderr._original_stdstream_copy
            sys.stderr._original_stdstream_copy = normal_stderr

    def __exit__(self, exc_type, exc_value, traceback):
        if self.restore_stdout is not None:
            sys.stderr._original_stdstream_copy = self.restore_stdout
        if self.restore_stderr is not None:
            sys.stderr._original_stdstream_copy = self.restore_stderr


def strip_multiline_string_indents(f):

    source = textwrap.dedent(inspect.getsource(f))
    blanks = (
        "\n" * f.__code__.co_firstlineno
    )  # padd the source so the lines in the file line up for the debugger
    source = blanks + "\n".join(
        source.splitlines()[1:]
    )  # remove the decorator first line.

    # define the external closure variables so f.__closure__ will match our recompiled version
    if len(f.__code__.co_freevars) > 0:
        raise Exception(
            "You currently must use @guidance(dedent=False) for closure functions (function nested within other functions that reference the outer functions variables)!"
        )
        lines = source.split("\n")
        lines[0] = "def __outer__closure_wrap():"
        lines[1] = (
            "    "
            + ",".join(f.__code__.co_freevars)
            + " = "
            + ",".join("None" for _ in f.__code__.co_freevars)
        )
        source = "    \n".join(
            lines
        )  # TODO: this does not quite work because new_code_obj is now the __outer__closure_wrap() function...could be fixed with work...

    old_code_obj = f.__code__
    old_ast = ast.parse(source)
    r = _Rewrite()
    r.source_lines = source.split("\n")
    r.indentation = [None for l in r.source_lines]
    r.start_counts = [0 for l in r.source_lines]
    # r._avoid_backslashes = True
    new_ast = r.visit(old_ast)
    new_code_obj = compile(new_ast, old_code_obj.co_filename, "exec")

    # find the code block
    for i in range(len(new_code_obj.co_consts)):
        if str(type(new_code_obj.co_consts[i])) == "<class 'code'>":
            break

    # create a new function based on the modified code
    new_f = types.FunctionType(
        new_code_obj.co_consts[i],
        f.__globals__,
        name=f.__name__,
        argdefs=f.__defaults__,
        closure=f.__closure__,
    )
    new_f.__kwdefaults__ = f.__kwdefaults__
    return new_f


class CaptureEvents:
    """Creates a scope where all the events are captured in a queue.

    Note that this does not stop the events from being captured by higher level scopes.
    """

    def __init__(self, lm):
        self.lm = lm

    def __enter__(self):
        self.lm._event_queue = queue.Queue()
        return self.lm._event_queue

    def __exit__(self, type, value, traceback):
        self.lm._event_queue = None


class JupyterComm:
    def __init__(
        self, target_id, ipython_handle, callback=None, on_open=None, mode="register"
    ):
        from ipykernel.comm import Comm

        self.target_name = "guidance_interface_target_" + target_id
        # print("TARGET NAME", self.target_name)
        self.callback = callback
        self.jcomm = None
        self.ipython_handle = ipython_handle
        self.addd = 1
        self.send_queue = asyncio.Queue()
        self.open_event = asyncio.Event()
        self.is_open = False
        asyncio.get_event_loop().create_task(self._send_loop())
        if mode == "register":
            # log("REGISTERING", self.target_name)
            # asyncio.get_event_loop().create_task(self._register())
            def comm_opened(comm, open_msg):
                # log("OPENED")
                self.addd = 2
                self.jcomm = comm
                self.is_open = True
                self.jcomm.on_msg(self._fire_callback)
                self.open_event.set()
                self._fire_callback({"content": {"data": {"event": "opened"}}})

            self.ipython_handle.kernel.comm_manager.register_target(
                self.target_name, comm_opened
            )
            # get_ipython().kernel.comm_manager.register_target(self.target_name, comm_opened) # noqa: F821
        elif mode == "open":
            # log("OPENING", self.target_name)
            self.jcomm = Comm(target_name=self.target_name)
            self.jcomm.on_msg(self._fire_callback)
            # self._fire_callback({"content": {"data": "opened"}})
        else:
            raise Exception("Passed mode must be either 'open' or 'register'!")

    def clear_send_queue(self):
        while not self.send_queue.empty():
            self.send_queue.get_nowait()
            self.send_queue.task_done()

    def _fire_callback(self, msg):
        self.callback(msg["content"]["data"])

    def send(self, data):
        self.send_queue.put_nowait(data)

    async def _send_loop(self):
        while True:
            # log("SENDING_LOOP")
            if self.jcomm is None:
                self.open_event.clear()
                await self.open_event.wait()
            data = await self.send_queue.get()
            # log("SENDING_LOOP got one!")
            self.jcomm.send({"data": json.dumps(data)})

    # async def _waiting_send(self, data):
    #     #log("SENDING", self.jcomm, data)

    #     # await the open event if needed
    #     if self.jcomm is None:
    #         self.open_event.clear()
    #         await self.open_event.wait()
    #     #log("SENDING_now", self.jcomm, data)
    #     self.jcomm.send({"data": json.dumps(data)}) # we encode the JSON so iPython doesn't mess it up


# https://stackoverflow.com/questions/15411967/how-can-i-check-if-code-is-executed-in-the-ipython-notebook
def is_interactive():
    import __main__ as main

    return not hasattr(main, "__file__")


def log_softmax(array: np.ndarray, axis: int = -1) -> np.ndarray:
    # https://docs.scipy.org/doc/scipy/reference/generated/scipy.special.log_softmax.html
    array_maxs: np.ndarray = np.amax(array, axis=axis, keepdims=True)
    if array_maxs.ndim > 0:
        array_maxs[~np.isfinite(array_maxs)] = 0
    elif not np.isfinite(array_maxs):
        array_maxs = np.zeros(array_maxs.shape)
    subtract_maxs = array - array_maxs
    exp = np.exp(subtract_maxs)
    # suppress warnings about log of zero
    with np.errstate(divide="ignore"):
        summed = np.sum(exp, axis=axis, keepdims=True)
        out = np.log(summed)
    return subtract_maxs - out


def softmax(array: np.ndarray, axis: int = -1) -> np.ndarray:
    # https://docs.scipy.org/doc/scipy/reference/generated/scipy.special.softmax.html
    array_maxs = np.amax(array, axis=axis, keepdims=True)
    exp_x_shifted = np.exp(array - array_maxs)
    return exp_x_shifted / np.sum(exp_x_shifted, axis=axis, keepdims=True)



import sys

# Check if 'setup.py' is run directly with 'build'
# TODO: Consider generalizing this check further?
if __name__ == "__main__":
    if len(sys.argv) >= 2 and sys.argv[1] == "build":
        raise SystemExit(
            "Error: Direct invocation of 'setup.py build' is not recommended."
            "Please use 'pip' to build and install this package, like so:\n"
            "  pip install . (for the current directory)\n"
            "  pip install -e . (for an editable install)\n"
            "  pip wheel . --no-deps (to build a wheel)"
        )

import os
import re
import codecs
from setuptools import setup, find_packages
from pybind11.setup_helpers import Pybind11Extension, build_ext

here = os.path.abspath(os.path.dirname(__file__))

install_requires = [
    "diskcache",
    "numpy",
    "ordered_set",
    "platformdirs",
    "pyformlang",
    "protobuf",
    "pydantic",
    "requests",
    "tiktoken>=0.3",
]

# Our basic list of 'extras'
extras_requires = {
    "azureai": ["openai>=1.0"],
    "openai": ["openai>=1.0"],
    "schemas": ["jsonschema"],
    "server": ["fastapi", "uvicorn"],
}

# Create the union of all our requirements
all_requires = set()
for v in extras_requires.values():
    all_requires = all_requires.union(v)

# Required for builds etc.
doc_requires = ["ipython", "nbsphinx", "numpydoc", "sphinx_rtd_theme", "sphinx"]
test_requires = [
    "jupyter",
    "papermill",
    "pytest",
    "pytest-cov",
    "torch",
    "transformers",
    "mypy==1.9.0",
    "types-protobuf",
    "types-regex",
    "types-requests",
    "types-jsonschema",
]


def read(*parts):
    with codecs.open(os.path.join(here, *parts), "r") as fp:
        return fp.read()


def find_version(*file_paths):
    version_file = read(*file_paths)
    version_match = re.search(r"^__version__ = ['\"]([^'\"]*)['\"]", version_file, re.M)
    if version_match:
        return version_match.group(1)
    raise RuntimeError("Unable to find version string.")


setup(
    name="guidance",
    version=find_version("guidance", "__init__.py"),
    url="https://github.com/guidance-ai/guidance",
    author="Guidance Maintainers",
    author_email="maintainers@guidance-ai.org",
    description="A guidance language for controlling large language models.",
    long_description="Guidance enables you to control modern language models more effectively and efficiently than traditional prompting or chaining. Guidance programs allow you to interleave generation, prompting, and logical control into a single continuous flow matching how the language model actually processes the text.",
    packages=find_packages(exclude=["notebooks", "client", "tests", "tests.*"]),
    package_data={"guidance": ["resources/*"]},
    ext_modules=[
        Pybind11Extension(
            "guidance.cpp", ["guidance/_cpp/main.cpp", "guidance/_cpp/byte_trie.cpp"]
        )
    ],
    cmdclass={"build_ext": build_ext},
    python_requires=">=3.8",
    install_requires=install_requires,
    extras_require={
        "all": all_requires,
        "docs": doc_requires,
        "test": test_requires,
        **extras_requires,
    },
)



from .._guidance import guidance
from .._grammar import byte_range


@guidance(stateless=True)
def any_char(lm):
    # TODO: extend this to support utf-8 encoded multibyte unicode characters
    return lm + byte_range(b"\x00", b"\xff")



import base64
import os

from typing import TYPE_CHECKING

try:
    import pydantic

    from fastapi import FastAPI, HTTPException, Security
    from fastapi.security import APIKeyHeader
    from fastapi.responses import StreamingResponse
except ImportError:
    if TYPE_CHECKING:
        raise

from .models._model import Model, Engine
from ._grammar import GrammarFunction


class GuidanceRequest(pydantic.BaseModel):
    parser: str = pydantic.Field(
        title="parser",
        description="The text generated so far by the guidance program",
    )
    grammar: str = pydantic.Field(
        title="grammar",
        description="Guidance grammar to constrain the next characters generated",
    )


class Server:

    def __init__(self, engine, api_key=None, ssl_certfile=None, ssl_keyfile=None):
        """This exposes an Engine object over the network."""

        if isinstance(engine, Model):
            engine = engine.engine
        elif not isinstance(engine, Engine):
            raise TypeError("engine must be an Engine object")
        self.engine = engine
        self.app = FastAPI()
        self.valid_api_keys = self._load_api_keys(api_key)
        if ssl_certfile is None:
            ssl_certfile = os.getenv("GUIDANCE_SSL_CERTFILE")
        if ssl_keyfile is None:
            ssl_keyfile = os.getenv("GUIDANCE_SSL_KEYFILE")
        self.ssl_certfile = ssl_certfile
        self.ssl_keyfile = ssl_keyfile

        api_key_header = APIKeyHeader(name="x-api-key", auto_error=False)

        # def get_api_key(api_key_header: str = Security(api_key_header)) -> str:
        #     if api_key_header in self.valid_api_keys:
        #         return api_key_header
        #     raise HTTPException(
        #         status_code=status.HTTP_401_UNAUTHORIZED,
        #         detail="Invalid or missing API Key",
        #     )

        @self.app.post("/extend")
        async def extend_parser(
            guidance_request: "GuidanceRequest",
            x_api_key: str = Security(api_key_header),
        ):
            if x_api_key not in self.valid_api_keys:
                raise HTTPException(status_code=401, detail="Invalid API key")

            # data = await request.json()
            # parser = data.get("parser")
            grammar = GrammarFunction.deserialize(
                base64.b64decode(guidance_request.grammar)
            )

            return StreamingResponse(
                self.engine(guidance_request.parser, grammar),
                media_type="application/json",
            )

    def _load_api_keys(self, api_key):
        valid_api_keys = set()
        if api_key is None:
            api_key = os.getenv("GUIDANCE_API_KEY")
            if api_key:
                valid_api_keys.add(api_key)
        else:
            valid_api_keys.add(api_key)
        return valid_api_keys

    def run(self, host="localhost", port=8000):
        # Use uvicorn or another ASGI server to run
        import uvicorn

        uvicorn.run(
            self.app,
            host=host,
            port=port,
            ssl_certfile=self.ssl_certfile,
            ssl_keyfile=self.ssl_keyfile,
        )  # use host="0.0.0.0" for remote access



from .._model import Instruct
from ._llama_cpp import LlamaCpp, LlamaCppChat


class MistralChat(LlamaCppChat):
    def get_role_start(self, role_name, **kwargs):
        if role_name == "user":
            return "[INST] "

        elif role_name == "assistant":
            return ""

        elif role_name == "system":
            raise Exception("MistralChat does not support a sytem role!")

    def get_role_end(self, role_name=None):
        if role_name == "user":
            return " [/INST]"
        elif role_name == "assistant":
            return "</s>"
        elif role_name == "system":
            raise Exception("MistralChat does not support a sytem role!")


class MistralInstruct(LlamaCpp, Instruct):
    def get_role_start(self, role_name, **kwargs):
        if role_name == "instruction":
            return "[INST] "

    def get_role_end(self, role_name=None):
        if role_name == "instruction":
            return " [/INST]"



from .._guidance import guidance
from .._grammar import byte_range, select, string, Byte
from ._zero_or_more import zero_or_more
from ._any_char_but import any_char_but
import re
from types import SimpleNamespace
from pyformlang.regular_expression import PythonRegex


@guidance(stateless=True, dedent=False)
def regex(lm, pattern):
    # find all of the brackets we'll need to negate later
    nots = re.findall(r"\[\^(.*?)\]", pattern)
    nots = [re.compile("[" + x + "]") for x in nots]
    # Compensating for a weird design choice in pyformlang where they don't accept \n in .
    pattern = re.sub(r"(?<!\\)\.", "(.|\n)", pattern)
    regex = PythonRegex(pattern)
    cfg = tree_to_grammar(simplify_tree(regex), nots=nots)
    return cfg


# This is just a helper class so I can merge nodes without having to worry about pyformlang types
class FakeNode:
    def __init__(self, value, sons):
        self.head = SimpleNamespace()
        self.head.value = value
        self.sons = sons

    def get_tree_str(self, depth: int = 0) -> str:
        """Get a string representation of the tree behind the regex"""
        temp = " " * depth + str(self.head) + "\n"
        for son in self.sons:
            temp += son.get_tree_str(depth + 1)
        return temp


#
def tree_to_grammar(node, nots):
    """Takes a pyformlang regex tree and returns a guidance gramar"""
    if not node.sons:
        val = node.head.value
        if val == "Epsilon" or val == "Empty":
            val = ""
        else:
            val = string(val)
        return val
    if node.head.value == "Union":
        vals = [tree_to_grammar(x, nots) for x in node.sons]
        # If select starts with ^ and is a select between bytes and it is listed in our negations, negate it. The way I implemented it is a major hack, should fix
        if all([isinstance(x, Byte) for x in vals]) and vals[0].byte == b"^":
            temp_vals = [x.byte for x in vals[1:]]
            all_chars = b"".join(temp_vals).decode("utf8")
            for notz in nots:
                # The not matches all of the chars.
                # TODO: Technically there could be a situation where this is true AND this is not actually what we had in a regex, so this code is wrong. But it's probably never going to happen, and we'll replace this code later.
                if all([notz.match(x) is not None for x in all_chars]):
                    return any_char_but(temp_vals)

            # if b''.join(temp_vals).decode('utf8') in nots:
            #     # print('Negating', temp_vals)
            #     return select([byte_range(x[0], x[1]) for x in negative_byte_range(temp_vals)])
        return select(vals)
    if node.head.value == "Concatenation":
        ret = ""
        for x in node.sons:
            ret += tree_to_grammar(x, nots)
        return ret
    if node.head.value == "Kleene Star":
        return zero_or_more(tree_to_grammar(node.sons[0], nots))


def simplify_tree(regex):
    """Merges sequence of byte-based concats or unions into single nodes, to make the grammar more compact"""
    if len(regex.sons) and regex.head.value in ["Concatenation", "Union"]:
        regex = merge_nodes(regex, regex.head.value)
    regex.sons = [simplify_tree(x) for x in regex.sons]
    return regex


def merge_nodes(regex, op):
    """Merges node sequences of concats and unions.
    op must be 'Concatenation' or 'Union'
    """
    assert op in ["Concatenation", "Union"]
    current = regex
    if len(current.sons) and current.head.value == op:
        val = []
        to_visit = [current]
        while to_visit:
            current = to_visit.pop()
            leaves = [x for x in current.sons if len(x.sons) == 0]
            ops = [x for x in current.sons if len(x.sons) and x.head.value == op]
            if len(leaves) + len(ops) == len(current.sons):
                val.extend([x.head.value for x in leaves])
                to_visit += [
                    x for x in current.sons if len(x.sons) and x.head.value == op
                ]
        # if all I have is leaves, I can group them
        if all([len(x.sons) == 0 for x in current.sons]) and current.head.value == op:
            # val.extend([x.head.value for x in current.sons])
            if op == "Concatenation":
                val = "".join(val)
                new_node = FakeNode(val, [])
            elif op == "Union":
                new_node = FakeNode("Union", [FakeNode(x, []) for x in val])
            return new_node
    if len(val) > 1:
        if op == "Concatenation":
            # merge all chars into a string, return a concat between the string (as a leaf) and whatever right-side children exist
            val = "".join(val)
            new_node = FakeNode(op, [FakeNode(val, []), merge_nodes(current, op)])
        elif op == "Union":
            # Merge all select options into a left leaf of a select, whatever operation is left becomes the right child.
            new_node = FakeNode(
                op,
                [
                    FakeNode("Union", [FakeNode(x, []) for x in val]),
                    merge_nodes(current, op),
                ],
            )
        return new_node
    else:
        return regex



class ByteTrie:
    """A python implementation mirroring the C++ ByteTrie class."""

    def __init__(self, byte_strings=None, values=None, parent=None):
        self._parent = parent
        self.match_version = -1
        self.match = False
        self.partial_match = False
        self.prob = 0
        self.value = -1
        self.children = {}

        if byte_strings is not None:
            if values is None:
                for s in byte_strings:
                    self.insert(s, 0)
            else:
                for i, s in enumerate(byte_strings):
                    self.insert(s, values[i])

    def keys(self):
        return self.children.keys()

    def has_child(self, byte):
        return byte in self.children

    def child(self, byte):
        return self.children[byte]

    def parent(self):
        return self._parent

    def size(self):
        return len(self.children)

    def __len__(self):
        return self.size()

    def insert(self, s, value, pos=0):
        if len(s) <= pos:
            if self.value < 0:
                self.value = value
        else:
            first_byte = s[pos : pos + 1]
            if first_byte not in self.children:
                self.children[first_byte] = ByteTrie(parent=self)
            self.children[first_byte].insert(s, value, pos + 1)

    def compute_probs(self, probs):
        self.prob = 0.0

        if self.value != -1:
            self.prob += probs[self.value]

        if self.children:
            for k in self.children:
                child = self.children[k]
                child.compute_probs(probs)
                self.prob += child.prob



from ._block import block


def monospace():
    return block(
        opener="<||_html:<span style='font-family: Menlo, Monaco, monospace; font-size: 13px;'>_||>",
        closer="<||_html:</span>_||>",
    )



import base64
import copy
import html
import logging
import queue
import re
import threading
import time


from pprint import pprint
from typing import Dict, TYPE_CHECKING


import numpy as np

try:
    from IPython.display import clear_output, display, HTML

    ipython_is_imported = True
except ImportError:
    ipython_is_imported = False
try:
    import torch

    torch_is_imported = True
except ImportError:
    torch_is_imported = False


logger = logging.getLogger(__name__)
try:
    from .. import cpp  # type: ignore[attr-defined]
except ImportError:
    logger.warn(
        "Failed to load guidance.cpp, falling back to Python mirror implementations..."
    )
    from .. import _cpp as cpp
from ._guidance_engine_metrics import GuidanceEngineMetrics
from .._utils import softmax, CaptureEvents
from .._parser import EarleyCommitParser, Parser
from .._grammar import (
    GrammarFunction,
    string,
    _call_pool,
    _tag_pattern,
    Null,
    replace_model_variables,
    unreplace_model_variables,
    select,
)

from .. import _serialization_pb2

if TYPE_CHECKING:
    from ..library._block import ContextBlock

# define some constants we will reuse many times
_null_grammar = string("")
format_pattern = re.compile(r"<\|\|_.*?_\|\|>", flags=re.DOTALL)
nodisp_pattern = re.compile(
    r"&lt;\|\|_#NODISP_\|\|&gt;.*?&lt;\|\|_/NODISP_\|\|&gt;", flags=re.DOTALL
)
html_pattern = re.compile(r"&lt;\|\|_html:(.*?)_\|\|&gt;", flags=re.DOTALL)
image_pattern = re.compile(r"&lt;\|_image:(.*?)\|&gt;")


class Tokenizer:
    """This is the standardized tokenizer interface used by guidance models.

    This class should be subclassed by specific implementations and then used as the
    tokenizer in the corresponding Engine subclass.
    """

    def __init__(self, tokens, bos_token_id=None, eos_token_id=None):

        # a numpy array of token byte strings indexed by their token id
        if isinstance(tokens, list):
            self.tokens = np.array(
                tokens, dtype="object"
            )  # note that we need np.bytes_ to zero bytes are not treated as null terminations

        # a numpy array of token byte strings indexed by their token id
        elif isinstance(tokens, np.ndarray):
            self.tokens = tokens

        else:
            raise Exception("Unknown tokenizer was passed!")

        assert isinstance(
            self.tokens[0], bytes
        ), "The tokens need to be provided as bytes!"

        self.bos_token_id = bos_token_id
        self.bos_token = (
            None if self.bos_token_id is None else self.tokens[self.bos_token_id]
        )
        self.eos_token_id = eos_token_id if eos_token_id is not None else bos_token_id
        self.eos_token = (
            None if self.eos_token_id is None else self.tokens[self.eos_token_id]
        )

        # track which tokens are duplicates
        self.duplicate_tokens = []
        found = {}
        for i, t in enumerate(self.tokens):
            if t in found:
                self.duplicate_tokens.append((i, found[t]))
            else:
                found[t] = i

    def __call__(self, byte_string):
        """Returns a list of tokens that represent the given byte string."""
        raise NotImplementedError(
            "You need to use a Tokenize subclass that overrides the __call__ method"
        )

    def clean_duplicate_tokens(self, probs):
        """This moves all the probability mass from duplicate positons on to their primary index."""
        for i, j in self.duplicate_tokens:
            probs[j] += probs[i]
            probs[i] = 0


class EngineCallResponse:
    new_bytes: bytes
    is_generated: bool
    new_bytes_prob: float
    capture_groups: dict
    capture_group_log_probs: dict
    new_token_count: int

    def __init__(
        self,
        new_bytes,
        is_generated,
        new_bytes_prob,
        capture_groups,
        capture_group_log_probs,
        new_token_count,
    ):
        self.new_bytes = new_bytes
        self.is_generated = is_generated
        self.new_bytes_prob = new_bytes_prob
        self.capture_groups = capture_groups
        self.capture_group_log_probs = capture_group_log_probs
        self.new_token_count = new_token_count

    def _to_proto(self):
        """Converts an EngineCallResponse object to its Protobuf representation.

        Returns:
            engine_response_pb2.EngineCallResponse: The Protobuf equivalent of this object.
        """
        return _serialization_pb2.EngineCallResponse(
            new_bytes=self.new_bytes,
            is_generated=self.is_generated,
            new_bytes_prob=self.new_bytes_prob,
            capture_groups=self.capture_groups,
            capture_group_log_probs=self.capture_group_log_probs,
            new_token_count=self.new_token_count,
        )

    def encode(self, charset):
        """Used to support FastAPI encoding of EngineCallResponse objects."""
        return self.serialize()

    def serialize(self):
        proto = self._to_proto()
        return proto.SerializeToString()

    @staticmethod
    def deserialize(byte_data):
        proto = _serialization_pb2.EngineCallResponse()
        proto.ParseFromString(byte_data)
        return EngineCallResponse(
            new_bytes=proto.new_bytes,
            is_generated=proto.is_generated,
            new_bytes_prob=proto.new_bytes_prob,
            capture_groups=proto.capture_groups,
            capture_group_log_probs=proto.capture_group_log_probs,
            new_token_count=proto.new_token_count,
        )


class Engine:
    """The engine owns the inference computation and is used/created by the Model class.

    Engine objects represent the expensive parts of inference. While Model objects are cheap and do not
    need to know about the tokenizer or the model parameters, Engine objects know about both. Many
    Model objects can reference a single Engine object. Engine objects can also be hidden behind a
    Server so a single server can serve many clients' model objects through a single Engine object.
    """

    def __init__(self, tokenizer, compute_log_probs=False):
        self.tokenizer = tokenizer
        self.compute_log_probs = compute_log_probs

        # build a prefix tree of the tokens
        self._token_trie = cpp.ByteTrie(
            self.tokenizer.tokens, np.arange(len(self.tokenizer.tokens))
        )
        self._token_trie.match = True
        self._token_trie.match_version = 0

        self.metrics = GuidanceEngineMetrics()

    def reset_metrics(self):
        self.metrics = GuidanceEngineMetrics()

    def start(self, parser, grammar, ensure_bos_token=True):
        """Start processing parser state executed through the grammar.

        Parameters
        ----------
        parser : str or Parser
            This is represents the current state of a guidance parser that will be extended
            using the passed grammar. If a string is given then we assume the previous parser
            state is just a fixed string prompt, if a full Parser is given then we extend that
            parser by appending the new grammar to the parser's current grammar and then
            inferencing the model. (TODO: implement full parser extension support)
        grammar: Grammar
            This is the grammar we are extending the parser with.
        """
        # def __call__(self, grammar, max_tokens=1000000, n=1, top_p=1, temperature=0.0, ensure_bos_token=True):
        # assert n == 1, "Still need to add support for n > 1!"

        # note we only support a fixed set of engine variables for the sake of security
        self._replacements = replace_model_variables(
            grammar, self, allowed_vars=["eos_token", "bos_token"]
        )

        # right now we only support a text/bytes prompt parser state, so we extract that
        if isinstance(parser, bytes):
            prompt = parser
        elif isinstance(parser, str):
            prompt = bytes(parser, encoding="utf8")
        elif isinstance(parser, Parser):
            raise NotImplementedError(
                "Still need to implement support for extending a full Parser state."
            )
        else:
            raise Exception("The passed parser is of an unknown type!")

        # add the beginning of sequence token if needed
        if (
            ensure_bos_token
            and self.tokenizer.bos_token is not None
            and not prompt.startswith(self.tokenizer.bos_token)
        ):
            prompt = self.tokenizer.bos_token + prompt

        # run a simple tokenizer (that does not use a grammar) on the prefix for better performance
        self._token_ids, self._token_byte_positions = self._tokenize_prefix(prompt)
        self._token_ids, self._token_byte_positions = self._cleanup_tokens(
            self._token_ids, self._token_byte_positions
        )
        if len(self._token_byte_positions) > 0:
            self._pre_parser_bytes = self._token_byte_positions[-1]
            self._trimmed_prompt_prefix = prompt[: self._token_byte_positions[-1]]
            prompt = prompt[self._token_byte_positions[-1] :]
        else:
            self._trimmed_prompt_prefix = b""
            self._pre_parser_bytes = 0

        # create a parser with a grammar that includes both our context and the passed grammar
        self._parser = EarleyCommitParser(prompt + grammar)

        # loop until we have generated a complete pattern
        self._hidden_count = len(prompt)  # we don't emit the prompt
        self._generated_pos = 0
        self._sampled_token_ind = None
        self._token_count = 0
        self._last_token_count = 0
        self._was_forced = False
        self._captured_data = {}
        self._captured_log_prob_data = {}

    def next(self, logits):
        """Move the grammar state machine processing forward to the next point where
            either get_logits is required to be called or we have a partial response
            to stream back.

        Parameters
        ----------
        logits : the logits obtained from the LLM after the last return from next(...)
        """

        logits_state = None
        response_state = None

        token_pos = 0
        is_generated = True

        is_new_token = False
        if logits is not None:
            is_new_token = True

            # if requested we compute the log probabilities so we can track the probabilities of each node
            if self.compute_log_probs:
                if torch_is_imported:
                    probs = (
                        torch.nn.functional.softmax(torch.tensor(logits), dim=-1)
                        .cpu()
                        .numpy()
                    )  # note we don't adjust for temp since we consider that a sampling step, not part of the probs
                else:
                    probs = softmax(
                        logits, axis=-1
                    )  # this numpy code is slower, so we don't use it if we have torch...
                self.tokenizer.clean_duplicate_tokens(probs)
                self._trie.compute_probs(probs)  # C++ impl
            else:
                probs = None

            grammar_temp = self._parser.next_byte_temperature()
            current_temp = grammar_temp if grammar_temp >= 0 else 0

            # get the sampling order
            if current_temp == 0:
                sampling_order = np.argsort(
                    -logits
                )  # we need numpy so the enumerate below does not get really slow...
            else:
                # assert top_p == 1, "Still need to add support for top_p!"
                if torch_is_imported:
                    logits = torch.tensor(logits)
                    torch.div(logits, current_temp, out=logits)
                    probs_torch = torch.nn.functional.softmax(logits, dim=-1)
                    sampling_order = (
                        torch.multinomial(probs_torch, len(probs_torch)).cpu().numpy()
                    )
                else:
                    # this numpy version allows us to drop our dependence on pytorch...but it is way slower
                    if probs is None:
                        probs = softmax(logits / current_temp, axis=-1)
                    probs += 1e-10  # ensure we have no zero probs that mess up numpy
                    probs /= np.sum(probs)
                    sampling_order = np.random.choice(
                        len(probs), size=len(probs), p=probs, replace=False
                    )  # the 1e-10 is ensure we have no zero probs, which numpy does not like

            # loop over the tokens looking for a valid one
            for i, self._sampled_token_ind in enumerate(sampling_order):
                self._sampled_token = self.tokenizer.tokens[self._sampled_token_ind]

                # break out if we have reach impossible tokens
                if logits[self._sampled_token_ind] <= -np.inf:
                    break

                # make sure it matches any forced prefix
                used_forced_pos = min(
                    self._forced_pos, self._start_pos + len(self._sampled_token)
                )
                if (
                    self._start_pos < self._forced_pos
                    and not self._sampled_token.startswith(
                        self._parser.bytes[self._start_pos : used_forced_pos]
                    )
                ):
                    continue
                offset = used_forced_pos - self._start_pos

                # make sure the parse is backed up to the position we want to start checking from TODO: make this account for shared prefixes with the last token
                self._parser.pos = used_forced_pos
                self._new_bytes_prob = 1.0

                # if we have gotten to the end of the valid tokens then we stop
                # if logits[self._sampled_token_ind] == -np.inf:
                #     raise self._report_failed_match(self._trimmed_prompt_prefix + self._parser.bytes)

                # check to see if the sampled token is allowed
                token_pos = offset
                node = (
                    self._trie
                )  # this is the Trie node we were left at when we could force the next byte above

                while token_pos < len(self._sampled_token):
                    next_byte = self._sampled_token[token_pos : token_pos + 1]
                    next_node = node.child(next_byte)

                    # if we don't have a cached match flag compute it using the grammar
                    if next_node.match_version < self._token_trie.match_version:
                        next_byte_mask = self._parser.next_byte_mask()
                        for (
                            byte
                        ) in (
                            node.keys()
                        ):  # we update all the children since the parser knows the full mask
                            child = node.child(byte)
                            child.match_version = self._token_trie.match_version
                            child.match = next_byte_mask[byte[0]]

                    # advance or fail according to the (now up-to-date) match cache
                    if next_node.match:

                        # get the parser to consume the next byte
                        if next_node.prob < 1e-8:
                            if node.prob < 1e-8:
                                log_prob_delta = 0
                            else:
                                log_prob_delta = -20
                        else:
                            log_prob_delta = np.log(next_node.prob) - np.log(node.prob)
                        # log_prob_delta = np.log(next_node.prob) - np.log(node.prob)
                        self._new_bytes_prob = next_node.prob
                        commit_point = self._parser.consume_byte(
                            next_byte, log_prob=log_prob_delta
                        )

                        # mark that we accepted this byte
                        node = next_node
                        token_pos += 1

                        # if we are at a hidden commit point then we need to hide the bytes that match that node
                        if commit_point is not None and commit_point.node.hidden:

                            # if we are capturing the data from this node we need to do that now since we are about to remove it
                            # TODO: build a whole parse tree under this commit_point node so we can record child node captures
                            if commit_point.node.capture_name:
                                self._captured_data[commit_point.node.capture_name] = (
                                    self._parser.bytes[commit_point.start :]
                                )
                                self._captured_log_prob_data[
                                    commit_point.node.capture_name
                                ] = commit_point.log_prob

                            # This takes the item and commits to it as part of the parse and then shrinks it to zero width
                            # in other words this hides the item
                            self._parser.commit_and_collapse_item(commit_point)

                            # keep the bytes we still need to emit
                            if self._forced_pos < commit_point.start:
                                self._parser.shadow_rewind(self._forced_pos)

                            else:
                                # pop off any tokens that overlap the hidden bytes
                                i = len(self._token_byte_positions) - 1
                                while (
                                    i >= 0
                                    and self._token_byte_positions[i]
                                    - self._pre_parser_bytes
                                    > commit_point.start
                                ):
                                    self._token_ids.pop()
                                    self._token_byte_positions.pop()
                                    self._token_count -= 1
                                    i -= 1
                                # re-add any bytes we cut too far on
                                self._parser.shadow_rewind(
                                    self._token_byte_positions[-1]
                                    - self._pre_parser_bytes
                                )
                            is_new_token = False
                            break

                        elif token_pos == len(self._sampled_token):
                            break  # this token is valid
                    else:
                        # partially valid tokens are okay if we are running off the end of a grammar, but not otherwise
                        if not self._parser.matched():
                            token_pos = -1

                        break  # this token is no longer valid

                # see if we are breaking out of the whole loop
                if not is_new_token:
                    break

                # check if this token is dominated by other longer valid tokens (and hence would never be consistent with greedy tokenization)
                # TODO: disabled for now because of sentencepeice non-local issues
                # if token_pos == len(self._sampled_token) and not self._parser.matched(): # not we don't check if we have matched, because then we can generate anything afterwards
                #     if _check_dominated(node, self._parser, self._token_trie.match_version, self._parser.next_byte_mask()):
                #         token_pos = -1

                if token_pos > 0:
                    break  # we found a valid token

                if self._parser.matched():
                    break  # if we already have a full match we don't try more tokens we just give up as soon as the model deviates from the grammar

        is_done = False
        while (
            True
        ):  # each iteration generates one more token (and some of the associated bytes)
            if is_new_token:
                # emit whatever we know will not be hidden
                new_bytes = self._parser.bytes[
                    self._generated_pos : self._parser.earliest_hidden_start()
                ]

                # if we cannot consume any more tokens then we are done
                if (
                    not self._is_forced
                    and token_pos < len(self._sampled_token)
                    and self._trie == self._token_trie
                ):

                    # which if can't consume any more tokens, but we are not yet done
                    if not self._parser.matched():
                        self._parser.matched()
                        raise self._report_failed_match(
                            self._trimmed_prompt_prefix + self._parser.bytes
                        )

                    # TODO: if we exactly match the end of the pattern then we can commit to this last token
                    # if m.span()[1] == len(generated_text):
                    #     self._cache_state["new_token_ids"].append(self._sampled_token_ind)

                    # capture the named groups from the parse tree
                    self._parser.get_captures(
                        self._captured_data, self._captured_log_prob_data
                    )

                    # we have no valid log prob data if we didn't compute it
                    # yield new_bytes[self._hidden_count:], self._is_generated, self._new_bytes_prob, self._captured_data, self._captured_log_prob_data, token_count - last_token_count

                    response_state = (
                        new_bytes[self._hidden_count :],
                        is_generated,
                        self._new_bytes_prob if self.compute_log_probs else 1.0,
                        self._captured_data,
                        self._captured_log_prob_data,
                        self._token_count - self._last_token_count,
                    )

                    self._last_token_count = self._token_count

                    # TODO: we only need to do this when we might re-use the grammar object...we might want to account for that
                    unreplace_model_variables(self._replacements)

                    is_done = True
                else:
                    self._generated_pos += len(new_bytes)

                    # yeild the snippet of text created by the next token
                    out = new_bytes[self._hidden_count :]
                    if len(out) > 0:
                        # capture the named groups from the (partial) parse tree, # TODO: disabled for now until we handle list_append correctly
                        # new_captured_data, new_captured_log_prob_data = self._parser.get_captures()
                        # self._captured_data.update(new_captured_data)
                        # self._captured_log_prob_data.update(new_captured_log_prob_data)
                        # yield out, self._is_generated, self._new_bytes_prob, self._captured_data, self._captured_log_prob_data, self._token_count - self._last_token_count # note that we don't capture groups until a complete parse right now...

                        response_state = (
                            out,
                            is_generated,
                            self._new_bytes_prob if self.compute_log_probs else 1.0,
                            self._captured_data,
                            self._captured_log_prob_data,
                            self._token_count - self._last_token_count,
                        )

                        self._last_token_count = self._token_count
                        self._hidden_count = 0
                        self._token_count += 1  # note we only update this for tokens that emit non-hidden content
                    else:
                        self._hidden_count -= len(new_bytes)

                    self._token_ids.append(self._sampled_token_ind)

                    # track the byte position of each token
                    if len(self._token_byte_positions) == 0:
                        self._token_byte_positions.append(len(self._sampled_token))
                    else:
                        self._token_byte_positions.append(
                            self._token_byte_positions[-1] + len(self._sampled_token)
                        )

                if response_state is not None:
                    break

            token_pos = 0
            is_generated = False

            is_new_token = True

            # note where we are starting for this token
            self._start_pos = self._parser.pos

            # let the parser know that we have advanced another token (used ofr tracking max token limits)
            self._parser.mark_new_token()

            # walk down the trie as far as possible before computing the logits
            self._trie = self._token_trie
            self._trie.match_version += (
                1  # this invalidates all the match caches from the previous token
            )
            # self._trie.prob = 0.0 # need to reset when we reset the match_version
            while True:
                next_byte_mask = self._parser.next_byte_mask()
                next_byte_mask_sum = next_byte_mask.sum()

                # see if we reached a dead end of the grammar
                if next_byte_mask_sum == 0:
                    break

                # if there is more than one option we cannot advance without computing the logits
                elif next_byte_mask_sum != 1:
                    break

                # we are not forced if we are at the end of the grammar
                elif self._parser.matched():
                    break

                # if there is only one possible next byte we can keep forcing
                elif next_byte_mask_sum == 1:

                    # look for valid children
                    next_byte = None
                    for byte in self._trie.keys():

                        # mark this self._trie node with an up-to-date match flag (may save work later)
                        node = self._trie.child(byte)
                        node.match_version = self._token_trie.match_version
                        # node.prob = 0.0 # reset when we reset the match_version
                        node.match = next_byte_mask[byte[0]]

                        # see if we found a match
                        if node.match:
                            next_byte = byte
                            break

                    # if we can't extend then this token is forced
                    if next_byte is None:
                        break

                    # otherwise since there is only one possible next byte we keep going
                    else:
                        commit_point = self._parser.consume_byte(
                            next_byte, log_prob=0.0
                        )

                        # if we are at a hidden commit point then we need to hide the bytes that match that node
                        if commit_point is not None and commit_point.node.hidden:

                            # This takes the item and commits to it as part of the parse and then shrinks it to zero width
                            # in other words this hides the item
                            self._parser.commit_and_collapse_item(commit_point)

                            # keep the bytes we still need to emit
                            if self._start_pos < commit_point.start:
                                self._parser.shadow_rewind(self._start_pos)

                            else:
                                # pop off any tokens that overlap the hidden bytes
                                i = len(self._token_byte_positions) - 1
                                while (
                                    i >= 0
                                    and self._token_byte_positions[i]
                                    - self._pre_parser_bytes
                                    > commit_point.start
                                ):
                                    self._token_ids.pop()
                                    self._token_byte_positions.pop()
                                    self._token_count -= 1
                                    i -= 1
                                # re-add any bytes we cut too far on
                                self._parser.shadow_rewind(
                                    self._token_byte_positions[-1]
                                    - self._pre_parser_bytes
                                )
                            is_new_token = False  # this restarts us at the top of the outer token gen loop
                            break

                        self._trie = self._trie.child(next_byte)

            self._forced_pos = self._parser.pos  # record how far the bytes are forced

            if is_new_token:
                # back up if we got forced up to a point that is not a valid token
                if next_byte_mask_sum <= 1:
                    while self._trie.value < 0 and self._trie.parent() is not None:
                        self._trie = self._trie.parent()
                        self._forced_pos -= 1
                    self._parser.pos = self._forced_pos

                # if we walked all the way to a forced token then we advance without computing the logits
                # we are forced if there are no more options and we are either in the middle of the grammar or at a trie leaf
                self._is_forced = next_byte_mask_sum <= 1 and (
                    len(self._trie) == 0
                    if self._parser.matched()
                    else self._trie != self._token_trie
                )
                if self._is_forced:
                    self._sampled_token_ind = self._trie.value
                    self._sampled_token = self.tokenizer.tokens[self._sampled_token_ind]
                    self._new_bytes_prob = 1.0
                    self._was_forced = True

                # we are at the end of the grammar
                elif next_byte_mask_sum == 0:

                    # mark the token we "sampled" if we have comsumed some bytes
                    if self._trie != self._token_trie:
                        self._sampled_token_ind = self._trie.value
                        self._sampled_token = self.tokenizer.tokens[
                            self._sampled_token_ind
                        ]
                        self._new_bytes_prob = 1.0

                # otherwise we need to compute the logits and sample a valid token
                else:

                    # if we were forced we might need to clean up the greedy tokenization to match the global tokenization behavior as seen in training
                    if self._was_forced:
                        self._token_ids, self._token_byte_positions = (
                            self._cleanup_tokens(
                                self._token_ids, self._token_byte_positions
                            )
                        )
                        self._was_forced = False

                    grammar_temp = self._parser.next_byte_temperature()
                    current_temp = grammar_temp if grammar_temp >= 0 else 0
                    logits_state = (
                        self._token_ids,
                        self._parser.bytes[self._start_pos : self._forced_pos],
                        current_temp,
                    )
                    break

        return is_done, logits_state, response_state

    def __call__(self, parser, grammar, ensure_bos_token=True):
        """Returns a new updated parser state executed through the grammar.

        Parameters
        ----------
        parser : str or Parser
            This is represents the current state of a guidance parser that will be extended
            using the passed grammar. If a string is given then we assume the previous parser
            state is just a fixed string prompt, if a full Parser is given then we extend that
            parser by appending the new grammar to the parser's current grammar and then
            inferencing the model. (TODO: implement full parser extension support)
        grammar: Grammar
            This is the grammar we are extending the parser with.
        """

        self.start(parser, grammar, ensure_bos_token)

        logits = None
        while True:
            is_done, logits_state, response_state = self.next(logits)
            logits = None

            if response_state is not None:
                (
                    response_new_bytes,
                    response_is_generated,
                    response_new_bytes_prob,
                    response_capture_groups,
                    response_capture_group_log_probs,
                    response_new_token_count,
                ) = response_state

                yield EngineCallResponse(
                    new_bytes=response_new_bytes,
                    is_generated=response_is_generated,
                    new_bytes_prob=response_new_bytes_prob,
                    capture_groups=response_capture_groups,
                    capture_group_log_probs=response_capture_group_log_probs,
                    new_token_count=response_new_token_count,
                )

            if logits_state is not None:
                token_ids, forced_bytes, current_temp = logits_state
                logits = self.get_logits(token_ids, forced_bytes, current_temp)

            if is_done:
                break

    def _tokenize_prefix(self, byte_string):
        """This is used to speed up the tokenization of long prompts without using the parser."""
        token_ids = []
        token_byte_positions = []

        # loop trying to decode a new token at each iteration
        pos = 0
        while True:

            # walk down the token trie looking for a unique token match
            trie = self._token_trie
            valid_pos = -1
            valid_value = -1
            while True:
                if pos >= len(byte_string):
                    if len(trie) > 0:
                        valid_pos = -1
                    break

                # check if we can keep going or are at a dead end
                if trie.has_child(byte_string[pos : pos + 1]):
                    trie = trie.child(byte_string[pos : pos + 1])
                    pos += 1

                    # record the last valid token down this path as we go
                    if trie.value >= 0:
                        valid_pos = pos
                        valid_value = trie.value
                else:
                    break  # we can't go any farther

            if valid_pos == -1:
                break
            else:
                token_ids.append(valid_value)
                token_byte_positions.append(valid_pos)
                pos = valid_pos

        return token_ids, token_byte_positions

    def _cleanup_tokens(self, token_ids, token_byte_positions):

        # compute a joint tokenization
        joint_token_ids = self._joint_tokenize(token_ids)

        # see if we need to redo the tokenization
        redo = False
        if len(joint_token_ids) != len(token_ids):
            redo = True
        else:
            for i, id in enumerate(joint_token_ids):
                if token_ids[i] != id:
                    redo = True
                    break

        if redo:
            token_ids = joint_token_ids
            last_pos = token_byte_positions[-1]
            token_byte_positions = []
            pos = 0
            for i, id in enumerate(joint_token_ids):
                pos += len(self.tokenizer.tokens[id])
                token_byte_positions.append(pos)

            # ugly hack to deal with sentence peice craziness of space hiding after special tokens TODO: figure out how to make this more robust
            if (
                token_byte_positions[-1] == last_pos + 1
                and self.tokenizer.tokens[token_ids[0]] == b"<s>"
                and self.tokenizer.tokens[token_ids[1]][0:1] == b" "
            ):
                for i in range(1, len(token_byte_positions)):
                    token_byte_positions[i] -= 1
            assert token_byte_positions[-1] == last_pos

        return token_ids, token_byte_positions

    def get_logits(self, token_ids, forced_bytes, current_temp):
        """A fake method designed to be overriden by subclasses."""

        # pretend to extend the KV cache and update the log probs
        return np.randn(len(self.tokenizer.tokens))

    def _report_failed_match(self, prompt):
        """Note that this can be overridden by subclasses that have more likely reasons than a bug in the token set (like remote models)."""
        return Exception(
            "We can't consume any more tokens, but we are not yet done! Perhaps your model's token set is incomplete? This happened after the prompt:"
            + str(prompt[-40:])
        )

    def _joint_tokenize(self, token_ids):
        """What a full joint tokenizer would give for a given byte string"""
        return token_ids


class Model:
    """The base guidance model object, which represents a model in a given state.

    Model objects are immutable representations of model state, so whenever you change
    them you get a new Model object. However, these copies share the "expensive"
    parts of the underlying model like the the parameters and KV-cache, through a shared
    Engine, so making copies of Model objects is cheap.

    .. automethod:: __add__
    """

    open_blocks: Dict["ContextBlock", None] = {}  # track what context blocks are open
    _grammar_only = 0  # a flag that tracks when we are forced to be executing only compiled grammars (like when we are inside a select)
    _throttle_refresh = 0  # a flag that tracks when we can throttle our display since we know future display calls are going to happen

    def __init__(self, engine, echo=True, **kwargs):
        """Build a new model object that represents a model in a given state.

        Note that this constructor is not meant to be used directly, since there

        Parameters
        ----------
        engine : Engine
            The inference engine to use for this model.
        echo : bool
            If true the final result of creating this model state will be displayed (as HTML in a notebook).
        """
        if isinstance(engine, str) and engine.startswith("http"):
            from ._remote import RemoteEngine

            engine = RemoteEngine(engine, **kwargs)

        # # auto-wrap the tokenizer in the standard guidance interface
        # if not isinstance(tokenizer, Tokenizer):
        #     tokenizer = Tokenizer(tokenizer)

        self.engine = engine
        self.echo = echo
        self.token_count = 0  # tracks how many tokens our byte state represents
        self.max_display_rate = 0.2  # this controls how frequently we are allowed to redraw the display (in seconds)
        self.opened_blocks = {}  # what context blocks have been opened but not closed
        # self.compute_log_probs = compute_log_probs

        # private attributes
        self._variables = {}  # these are the state variables stored with the model
        self._variables_log_probs = (
            {}
        )  # these are the state variables stored with the model
        self._cache_state = {}  # mutable caching state used to save computation
        self._state = ""  # the current bytes that represent the state of the model
        self._event_queue = None  # TODO: these are for streaming results in code, but that needs implemented
        self._event_parent = None
        self._last_display = (
            0  # used to track the last display call to enable throttling
        )
        self._last_event_stream = (
            0  # used to track the last event streaming call to enable throttling
        )

    @property
    def active_role_end(self):
        """The default end patterns we should use for `gen` calls.
        TODO: move this logic into the gen call...we can do with if we allow model_variables to run functions.

        These patterns are computed dynamically by the model object because they can depend on
        what the current open roles are, which is something
        """

        # add any active non-empty role ends. Ignore role ends that are spaces
        parts = []
        for _, role_end_str in self.opened_blocks.values():
            role_end_str = format_pattern.sub("", role_end_str)
            if len(role_end_str) > 0 and not re.fullmatch(r"\s+", role_end_str):
                parts.append(role_end_str)

        return select(parts)

    def _html(self):
        """Generate HTML that displays the model object."""
        display_out = self._state
        for context in reversed(self.opened_blocks):
            display_out += self.opened_blocks[context][1]
        display_out = html.escape(display_out)
        display_out = nodisp_pattern.sub("", display_out)
        display_out = html_pattern.sub(lambda x: html.unescape(x.group(1)), display_out)
        display_out = image_pattern.sub(
            lambda x: '<img src="data:image/png;base64,'
            + base64.b64encode(self[x.groups(1)[0]]).decode()
            + '" style="max-width: 400px; vertical-align: middle; margin: 4px;">',
            display_out,
        )
        display_out = (
            "<pre style='margin: 0px; padding: 0px; vertical-align: middle; padding-left: 8px; margin-left: -8px; border-radius: 0px; border-left: 1px solid rgba(127, 127, 127, 0.2); white-space: pre-wrap; font-family: ColfaxAI, Arial; font-size: 15px; line-height: 23px;'>"
            + display_out
            + "</pre>"
        )
        return display_out

    def _send_to_event_queue(self, value):
        """For streaming in code.

        TODO: Is this still needed?"""
        if self._event_queue is not None:
            self._event_queue.put(value)
        if self._event_parent is not None:
            self._event_parent._send_to_event_queue(value)

    def stream(self):
        return ModelStream(self)

    def copy(self):
        """Create a shallow copy of the model object."""

        # start with a shallow copy
        new_lm = copy.copy(self)

        # then copy a few things we need deeper copies of
        new_lm._variables = self._variables.copy()
        new_lm._variables_log_probs = self._variables_log_probs.copy()
        new_lm.opened_blocks = self.opened_blocks.copy()

        # create a new clean event queue
        new_lm._event_queue = (
            None  # we start with no event queue because nobody is listening to us yet
        )
        if self._event_queue is not None:
            new_lm._event_parent = (
                self  # the current lm has an event que we make it our parent
            )
        elif self._event_parent is not None:
            new_lm._event_parent = (
                self._event_parent
            )  # otherwise if the current event que has an event parent then that is also our parent

        return new_lm

    def _inplace_append(self, value, force_silent=False):
        """This is the base way to add content to the current LM object that is being constructed.

        All updates to the model state should eventually use this function.
        Note this should only be used after making a copy, otherwise immutability would be violated.

        Parameters
        ----------
        value : bytes
            The bytes we should append to our current state.
        """

        # update the byte state
        self._state += str(value)  # TODO: make _state to be bytes not a string

        # see if we should update the display
        if not force_silent:
            self._update_display()

        # this is for programmatic streaming among other things
        if Model._throttle_refresh > 0:
            curr_time = time.time()
            if curr_time - self._last_event_stream >= self.max_display_rate:
                self._last_event_stream = curr_time
                self._send_to_event_queue(self)
        else:
            self._send_to_event_queue(self)

    def _update_display(self, throttle=True):
        if self.echo:
            if Model._throttle_refresh > 0:
                curr_time = time.time()
                if throttle and curr_time - self._last_display < self.max_display_rate:
                    return  # we are throttling the update
                else:
                    self._last_display = curr_time

            if ipython_is_imported:
                clear_output(wait=True)
                display(HTML(self._html()))
            else:
                pprint(self._state)

    def reset(self, clear_variables=True):
        """This resets the state of the model object.

        Parameters
        ----------
        clear_variables : bool
            If we should clear all the model object's variables in addition to reseting the byte state.
        """
        self._state = self._state[:0]
        if clear_variables:
            self._variables = {}
            self._variables_log_probs = {}
        return self

    def _repr_html_(self):
        if ipython_is_imported:
            clear_output(wait=True)
        return self._html()

    def _current_prompt(self):
        """The current prompt in bytes (which is the state without the context close tags)."""
        return format_pattern.sub("", self._state)

    def __str__(self):
        """A string representation of the current model object (that includes context closers)."""
        out = self._current_prompt()
        for context in reversed(self.opened_blocks):
            out += format_pattern.sub("", self.opened_blocks[context][1])
        return out

    def __add__(self, value):
        """Adding is the primary mechanism for extending model state.

        Parameters
        ----------
        value : guidance grammar
            The grammar used to extend the current model.
        """

        # create the new lm object we will return
        # (we need to do this since Model objects are immutable)
        lm = self.copy()

        # inside this context we are free to drop display calls that come too close together
        with throttle_refresh():

            # find what new blocks need to be applied
            new_blocks = []
            for context in Model.open_blocks:
                if context not in lm.opened_blocks:
                    new_blocks.append(context)

                    # mark this so we don't re-add when computing the opener or closer (even though we don't know the close text yet)
                    lm.opened_blocks[context] = (0, "")

            # find what old blocks need to be removed
            old_blocks = []
            for context in list(reversed(lm.opened_blocks)):
                if context not in Model.open_blocks and context in lm.opened_blocks:
                    old_blocks.append((lm.opened_blocks[context], context))

                    # delete this so we don't re-close when computing the opener or closer
                    del lm.opened_blocks[context]

            # close any newly closed contexts
            for (pos, close_text), context in old_blocks:
                if context.name is not None:
                    lm._variables[context.name] = format_pattern.sub(
                        "", lm._state[pos:]
                    )
                lm += context.closer

            # apply any newly opened contexts (new from this object's perspective)
            for context in new_blocks:
                lm += context.opener
                with grammar_only():
                    tmp = lm + context.closer
                close_text = tmp._state[
                    len(lm._state) :
                ]  # get the new state added by calling the closer
                lm.opened_blocks[context] = (len(lm._state), close_text)

                # clear out names that we override
                if context.name is not None:
                    if context.name in lm._variables:
                        del lm._variables[context.name]
                        if context.name in lm._variables_log_probs:
                            del lm._variables_log_probs[context.name]

            # wrap raw string values
            if isinstance(value, str):
                is_id = False
                parts = re.split(_tag_pattern, value)

                # we have no embedded objects
                if len(parts) == 1:
                    lm._inplace_append(value)
                    out = lm

                # if we have embedded objects we have to convert the string to a grammar tree
                else:
                    partial_grammar = _null_grammar
                    lm.suffix = ""
                    for i, part in enumerate(parts):
                        if i < len(parts) - 1:
                            lm.suffix = parts[i + 1]
                        if is_id:
                            call = _call_pool[part]
                            if isinstance(call, GrammarFunction):
                                partial_grammar += _call_pool[part]
                            else:
                                lm += partial_grammar
                                lm = _call_pool[part](lm)
                                partial_grammar = _null_grammar
                        elif part != "":
                            partial_grammar += string(part)
                        is_id = not is_id
                    out = lm + partial_grammar

            # if we find a null value we do nothing
            elif isinstance(value, Null):
                out = lm

            # run stateless functions (grammar nodes)
            elif isinstance(value, GrammarFunction):
                out = lm._run_stateless(value)

            # run stateful functions
            else:
                out = value(lm)
                if out is None:
                    raise Exception(
                        f"A guidance function returned `None`, not a model object! Did you forget to return the new lm at the end of your function?"
                    )
                if not isinstance(out, Model):
                    raise Exception(
                        f"A guidance function did not return a model object! Did you try to add a function to a model without calling the function? For example `model + guidance_function()` is correct, while `model + guidance_function` will cause this error."
                    )

        # this flushes the display
        out._inplace_append("")

        return out

    # def endswith(self, s):
    #     '''Checks if the current model state ends with the given value.'''
    #     return self._current_prompt().endswith(s)

    def __len__(self):
        """The string length of the current state.

        TODO: This should change to the byte length...
        """
        return len(str(self))

    def __setitem__(self, key, value):
        raise Exception(
            "Model objects are immutable so you can't use __setitem__! Consider using the .set(key, value) method instead to create a new updated model object."
        )

    def __getitem__(self, key):
        if key in self._variables:
            return self._variables[key]

        # look for named blocks that are still open with the given key as their name
        else:
            for context in list(reversed(self.opened_blocks)):
                if context.name == key:
                    return format_pattern.sub(
                        "", self._state[self.opened_blocks[context][0] :]
                    )

        raise KeyError(f"Model does not contain the variable '{key}'")

    def __contains__(self, item):
        return item in self._variables

    def get(self, key, default=None):
        """Return the value of a variable, or a default value if the variable is not present.

        Parameters
        ----------
        key : str
            The name of the variable.
        default : any
            The value to return if the variable is not current set.
        """
        return self._variables.get(key, default)

    def setattr(self, key, value):
        """Return a new model with the given model attribute set.

        Parameters
        ----------
        key : str
            The name of the attribute to be set.
        value : any
            The value to set the attribute to.
        """
        copy = self.copy()
        setattr(copy, key, value)
        return copy

    def delattr(self, key):
        """Return a new model with the given attribute deleted.

        Parameters
        ----------
        key : str
            The attribute name to remove.
        """
        copy = self.copy()
        delattr(copy, key)
        return copy

    def set(self, key, value):
        """Return a new model with the given variable value set.

        Parameters
        ----------
        key : str
            The name of the variable to be set.
        value : any
            The value to set the variable to.
        """
        copy = self.copy()
        copy._variables[key] = value
        return copy

    def remove(self, key):
        """Return a new model with the given variable deleted.

        Parameters
        ----------
        key : str
            The variable name to remove.
        """
        if key in self._variables:
            copy = self.copy()
            del copy._variables[key]
            if key in copy._variables_log_probs:
                del copy._variables_log_probs[key]
        else:
            copy = self
        return copy

    def log_prob(self, key, default=None):
        """Return the log prob of a variable, or a default value if the variable is not present.

        Parameters
        ----------
        key : str
            The name of the variable.
        default : any
            The value to return if the variable is not current set.
        """
        # TODO: support calling without a key to get the log prob of the whole model
        return self._variables_log_probs.get(key, default)

    # def get_cache(self):
    #     return self.engine.cache

    #     def tool_def(self, functions):

    #         self += """
    # # Tools

    # """
    #         if len(functions) > 0:
    #             self += '''## functions

    # namespace functions {

    # '''
    #         for function in functions:
    #             self += f"""// {function['description']}
    # type {function['name']} = (_: {{"""
    #             for prop_name,prop_data in function["parameters"]["properties"].items():
    #                 if "description" in prop_data:
    #                     self += f"\n// {prop_data['description']}\n"
    #                 self += prop_name
    #                 if prop_name not in function["parameters"]["required"]:
    #                     self += "?"
    #                 self += ": "
    #                 if "enum" in prop_data:
    #                     for enum in prop_data["enum"]:
    #                         self += f'"{enum}"'
    #                         if enum != prop_data["enum"][-1]:
    #                             self += " | "
    #                 else:
    #                     self += prop_data["type"]

    #                 if prop_name != list(function["parameters"]["properties"].keys())[-1]:
    #                     self += ",\n"
    #             self += """
    # }) => any;

    # """
    #             self[function['name']] = function
    #         self += "} // namespace functions\n"

    #         return self

    def _run_stateless(self, stateless_function, temperature=0.0, top_p=1.0, n=1):
        assert (
            Model._grammar_only == 0
        ), "We can't run grammar parsing while in context free mode! (for example inside a block closer)"

        logger.debug("start Model._run_stateless")

        # This needs to be here for streaming
        # if name is not None:
        #     self[name] = ""

        # replace ModelVariables with their actual values (note we save what we replaced so we can restore it later)
        replacements = replace_model_variables(stateless_function, self)

        # start the generation stream
        gen_obj = self.engine(self._current_prompt(), stateless_function)

        # we will return a new extended version of ourselves, which we track as `lm`
        lm = self

        # single generation
        if n == 1:
            generated_value = ""
            # logprobs_out = []

            delayed_bytes = b""
            # last_is_generated = False
            for chunk in gen_obj:

                # we make everything full probability if we are not computing uncertainty
                # if not self.engine.compute_log_probs:
                #     chunk.new_bytes_prob = 1.0

                # convert the bytes to a string (delaying if we don't yet have a valid unicode string)
                lm.token_count += chunk.new_token_count
                chunk.new_bytes = delayed_bytes + chunk.new_bytes
                try:
                    new_text = chunk.new_bytes.decode("utf8")
                except UnicodeDecodeError:
                    delayed_bytes = chunk.new_bytes
                    continue
                delayed_bytes = b""

                if len(chunk.new_bytes) > 0:
                    generated_value += new_text
                    if chunk.is_generated:
                        lm += f"<||_html:<span style='background-color: rgba({165*(1-chunk.new_bytes_prob) + 0}, {165*chunk.new_bytes_prob + 0}, 0, {0.15}); border-radius: 3px;' title='{chunk.new_bytes_prob}'>_||>"
                    lm += new_text
                    if chunk.is_generated:
                        lm += "<||_html:</span>_||>"

                # last_is_generated = chunk.is_generated

                if len(chunk.capture_groups) > 0:
                    for k in chunk.capture_groups:
                        v = chunk.capture_groups[k]

                        # see if we are in a list_append mode
                        if isinstance(v, list):
                            for i, inner_v in enumerate(v):
                                # convert to a string if possible
                                # TODO: will need to not just always do this once we support images etc.
                                try:
                                    inner_v = (
                                        inner_v.decode("utf8")
                                        if isinstance(inner_v, bytes)
                                        else inner_v
                                    )
                                except UnicodeDecodeError:
                                    pass

                                if k not in lm or not isinstance(
                                    lm._variables[k], list
                                ):
                                    lm._variables[k] = []
                                    lm._variables_log_probs[k] = []
                                lm._variables[k].append(inner_v)
                                lm._variables_log_probs[k].append(
                                    chunk.capture_group_log_probs[k][i]
                                )

                        # ...or standard assignment mode
                        else:
                            # convert to a string if possible
                            # TODO: will need to not just always do this once we support images etc.
                            try:
                                v = v.decode("utf8") if isinstance(v, bytes) else v
                            except UnicodeDecodeError:
                                pass
                            lm._variables[k] = v
                            lm._variables_log_probs[k] = chunk.capture_group_log_probs[
                                k
                            ]

            # if len(chunk.capture_groups) > 0:
            #     for k in chunk.capture_groups:
            #         v = chunk.capture_groups[k]
            #         lm[k] = v.decode("utf8") if isinstance(v, bytes) else v

        unreplace_model_variables(replacements)

        logger.debug("finish Model._run_stateless")

        return lm


class ModelStream:
    def __init__(self, model, grammar=None, timeout=5):
        """Create a model stream object that delays execution until it is iterated over."""
        if model.echo:
            model = model.copy()
            model.echo = False  # turn off display echoing
        self.model = model
        self.grammar = grammar
        self.timeout = timeout

    def __add__(self, grammar):
        """Extend this delayed chain of execution with another grammar append."""
        if self.grammar is None:
            return ModelStream(self.model, grammar)
        else:
            return ModelStream(self.model, self.grammar + grammar)

    def _inner_run(self, model):
        """This runs the model stream without iterating, and is only using internally by __iter__."""
        if isinstance(self.grammar, ModelStream):
            model = self.grammar._inner_run(model)
        elif self.grammar is None:
            model = self.model + ""
        else:
            model = self.model + self.grammar

    def __iter__(self):
        """Starts a thread to execute the model and grammar, yielding events as they occur."""

        # Create a thread-safe queue to hold events
        with CaptureEvents(self.model) as events:

            # Define the target function for the thread
            def target():
                try:
                    self._inner_run(self.model)
                    events.put(None)  # mark that we are done
                except BaseException as ex:
                    events.put(ex)

            # Start the thread
            thread = threading.Thread(target=target)
            thread.start()

            # Yield events from the queue as they become available
            while True:
                try:
                    # Wait for an event with a timeout to allow for thread termination
                    event = events.get(timeout=self.timeout)
                    if event is None:
                        break
                    elif isinstance(event, BaseException):
                        raise event
                    yield event
                except queue.Empty:
                    # Check if the thread is still alive
                    if not thread.is_alive():
                        break

            # Ensure the thread has completed
            thread.join()


class Chat(Model):
    """The base class for all chat-tuned models."""

    def get_role_start(self, role_name, **kwargs):
        """The starting grammar for a role.

        By default we follow the GPT role tag start conventions.

        Parameters
        ----------
        role_name : str
            The name of the role, like "user", or "assistant"
        kwargs : dict
            This kwargs are added to the role start as arguments.
        """
        return (
            "<|im_start|>"
            + role_name
            + "".join([f' {k}="{v}"' for k, v in kwargs.items()])
            + "\n"
        )

    def get_role_end(self, role_name=None):
        """The ending bytes for a role.

        Note that we cannot use a grammar in closers because they need to remain constant
        so we can append them whenever we need a representation before the final closing of the context.
        By default we follow the GPT role tag end conventions.

        Parameters
        ----------
        role_name : str
            The name of the role, like "user", or "assistant"
        """
        return "<|im_end|>"


class Instruct(Model):
    """The base class for all instruction-tuned models."""

    def get_role_start(self, role_name, **kwargs):
        raise Exception("Subclasses need to define what the role start should be!")

    def get_role_end(self, role_name=None):
        raise Exception("Subclasses need to define what the role end should be!")


class GrammarOnly:
    def __enter__(self):
        Model._grammar_only += 1

    def __exit__(self, exc_type, exc_value, traceback):
        Model._grammar_only -= 1


def grammar_only():
    """Returns a context manager that ensures only grammars are executed (not full python functions)."""
    return GrammarOnly()


class ThrottleRefresh:
    def __enter__(self):
        Model._throttle_refresh += 1

    def __exit__(self, exc_type, exc_value, traceback):
        Model._throttle_refresh -= 1


def throttle_refresh():
    """Returns a context manager that allows the print statement to drop display calls above the throttle rate."""
    return ThrottleRefresh()


class ConstraintException(Exception):
    pass


# def _compute_probs(trie, probs, found):
#     '''Computes the log probabilities for each internal trie node.'''
#     if trie.value is not None:
#         found[trie.value] = 1
#         trie.prob += probs[trie.value]

#     if len(trie) > 0:
#         # child_probs = []
#         for b in trie.keys():
#             child = trie.child(b)
#             _compute_probs(child, probs, found)
#             trie.prob += child.prob
#         # trie.log_prob = np.logaddexp.reduce(child_log_probs)


def _check_dominated(node, parser, match_version, next_byte_mask):
    curr_pos = parser.pos
    for byte_num in next_byte_mask.nonzero()[0]:
        next_byte = bytes((byte_num,))
        if not node.has_child(next_byte):
            return False  # no possible exension this direction, so we are not dominated
        child = node.child(next_byte)
        if child.match_version < match_version:
            child.match_version = match_version
            child.match = next_byte_mask[next_byte[0]]

        if not child.match:
            return False  # this child does not dominate the node, so the node is not dominated
        elif child.value is None:  # this child might not dominate the node
            parser.consume_byte(next_byte, log_prob=0.0)
            child_dominate = _check_dominated(
                child, parser, match_version, parser.next_byte_mask()
            )
            parser.pos = curr_pos
            if not child_dominate:
                return False
    return True